

#ifndef NULL
  #define  NULL    ((void *)0)
#endif


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Stride': Generalises the idea of an array index stride.      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class Stride {
protected :
  int base_v   ;
  int stride_v ;
  int norm_h   ;
  int stride_h ;

  int stodge ;  // bodge
public :
  Stride() ;
  Stride(const Stride&) ;

  int init(Range*, int) ;
  int step(Range*) ;
  int offset(Subscript*) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Frame': Representation of a general array entity.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


//class Vector ;

template<class T>
  class Array1;

class Frame {
  friend class Dispatch ;
  friend class Promotion ;
protected :
  int typeLen ;

  ProcessGroup* procGrp ;

  int rank ;
  Range**  rngs ;
  Stride** strs ;

  char* data ;

  //friend class Vector ;
  friend void copy(Frame*, Frame*) ;
  friend void remap(Frame*, Frame*) ;
  friend void vecGather(Frame*, Frame*, Array1<int>**) ;
  friend void vecScatter(Frame*, Frame*, Array1<int>**) ;
  friend void shift(Frame*, Frame*, const int, const int, const int mode) ;
  friend void skew(Frame* result, Frame* arr, Frame* shift,
                   const int dim, const int mode) ;
public :
  Frame(const int, ProcessGroup*, const int) ;
  Frame(Frame*, Subscript**, const int) ;
  ~Frame() ;

  // Adhoc-level inquiry functions
  Range* rng(const int i) {return rngs [i - 1] ; }
  ProcessGroup* prc() {return procGrp ; }

  // Lower-level access to data representation
  int rnk() {return rank ; }
  int len() {return typeLen ; }
  char* dat() {return data ; }
  Stride* str(const int i) {return strs [i - 1] ; }
  Range** shp() {return rngs ; }

  // Simple subscripting
  char* subs(Subscript**) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Data': Representation of an array variable.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class Data : public Frame {
public :
  Data(const int, ProcessGroup*, const int, Range**) ;
  ~Data() ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Promotion': an extension of an array, with added ranges      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


class Promotion : public Frame {
public :
  Promotion(const int _rank, Range** _rngs, Frame* parent, int* injection) ;
  Promotion(Frame* parent, int dim, Range* x) ;
  Promotion(const int, ProcessGroup*, const int, Range**, char*) ;
  ~Promotion() ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ArrayD': Type-secure interfaces to `Data'                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


template<class T>
class Array : public Data {
protected :
  Array(ProcessGroup& procGrp, const int rank, Range** rngs) :
      Data(sizeof(T), &procGrp, rank, rngs) {}
} ;


template<class T>
class Array0 : public Array<T> {
public :
  Array0(ProcessGroup& procGrp) :
      Array<T>(procGrp, 0, NULL) {}
  Array0() :
      Array<T>(*apg, 0, NULL) {}

  T& operator()() {
    return *((T*) subs(NULL)) ;
  }
} ;


template<class T>
class Array1 : public Array<T> {
public :
  Array1(ProcessGroup& procGrp, Range& x_1) :
      Array<T>(procGrp, 1, seq(&x_1)) {}
  Array1(Range& x_1) :
      Array<T>(*apg, 1, seq(&x_1)) {}

  T& operator()(Subscript& i_1) {
    return *((T*) subs(seq(&i_1))) ;
  }
} ;

template<class T>
class Array2 : public Array<T> {
public :
  Array2(ProcessGroup& procGrp, Range& x_1, Range& x_2) :
      Array<T>(procGrp, 2, seq(&x_1, &x_2)) {}
  Array2(Range& x_1, Range& x_2) :
      Array<T>(*apg, 2, seq(&x_1, &x_2)) {}

  T& operator()(Subscript& i_1, Subscript& i_2) {
    return *((T*) subs(seq(&i_1, &i_2))) ;
  }
} ;

template<class T>
class Array3 : public Array<T> {
public :
  Array3(ProcessGroup& procGrp, Range& x_1, Range& x_2, Range& x_3) :
      Array<T>(procGrp, 3, seq(&x_1, &x_2, &x_3)) {}
  Array3(Range& x_1, Range& x_2, Range& x_3) :
      Array<T>(*apg, 3, seq(&x_1, &x_2, &x_3)) {}

  T& operator()(Subscript& i_1, Subscript& i_2, Subscript& i_3) {
    return *((T*) subs(seq(&i_1, &i_2, &i_3))) ;
  }
} ;



inline Seq1<Subscript*> sect() {
  return seq((Subscript *) NULL) ;
}

inline Seq2<Subscript*> sect(Subscript& s1) {
  return seq(&s1, (Subscript *) NULL) ;
}

inline Seq3<Subscript*> sect(Subscript& s1, Subscript& s2) {
  return seq(&s1, &s2, (Subscript *) NULL) ;
}

inline Seq4<Subscript*> sect(Subscript& s1, Subscript& s2, Subscript& s3) {
  return seq(&s1, &s2, &s3, (Subscript *) NULL) ;
}


template<class T>
class Section1 : public Frame {
public :
  Section1(Array<T>& a, Subscript** ss) : Frame(&a, ss, 1) {}

  T& operator()(Subscript& i_1) {
    return *((T*) subs(seq(&i_1))) ;
  }
} ;


template<class T>
class Injection1 : public Promotion {
public :
  Injection1(Range& x_1, Array<T>& a, int* injection) :
      Promotion(1, seq(&x_1), &a, injection) {}

  T& operator()(Subscript& i_1) {
    return *((T*) subs(seq(&i_1))) ;
  }
} ;

template<class T>
class Injection2 : public Promotion {
public :
  Injection2(Range& x_1, Range& x_2, Array<T>& a, int* injection) :
      Promotion(2, seq(&x_1, &x_2), &a, injection) {}
  Injection2(Array1<T>& a, int dim, Range& x) : Promotion(&a, dim, &x) {}

  T& operator()(Subscript& i_1, Subscript& i_2) {
    return *((T*) subs(seq(&i_1, &i_2))) ;
  }
} ;


// ---------------------------------------------------------------
//
// Things to do:
//
//    Understand why `Promotion' has to be a friend of `Frame'
//    (the constructor uses protected fields of `Frame'---should
//    be ok, but doesn't work with some versions of g++).


